/*-
 * Copyright (c) 2008 Ivan Voras <ivoras@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "lib.h"
#include <err.h>
#include <paths.h>


/**
 * Creates a canonical transaction id string from the given numerical
 * transaction id string.
 */
void
make_trans_id_str(char *trans_str, int tid)
{
	snprintf(trans_str, MAX_TID_SIZE, "%0*d", MAX_TID_SIZE-1, tid);
}


/**	
 * Converts transaction ID to the canonical format (i.e. 000000004).
 * Garbage in, garbage out. A canonical transaction ID is 9 numerical
 * characters in length and begins with a 0. This supports about a
 * billion transactions.
 */
void
canonical_trans_id(char *dest, const char *raw_tid)
{
	int tid;
	tid = atoi(raw_tid);
	make_trans_id_str(dest, tid);
}


/**
 * Returns ID of last known transaction.
 */
void
get_last_trans_id(char *trans_id)
{
	DIR *d;
	struct dirent *dp;
	int max_tid;
	
	d = opendir(TRANS_DB_DIR);
	if (d == NULL)
		err(1, "Cannot open transaction database directory "
			TRANS_DB_DIR);
	max_tid = 0;
	while ((dp = readdir(d)) != NULL) {
		if (dp->d_name[0] != '0')
			continue;
		if (atoi(dp->d_name) > max_tid)
			max_tid = atoi(dp->d_name);
	}
	closedir(d);
	make_trans_id_str(trans_id, max_tid);
}


/**
 * Begins a new transaction (calls the pkg_trans utility) and fills in the
 * trans_id argument.
 * Returns 0 if successful. Nulls the trans_id argument if an error occurs. 
 */
int
trans_begin(char *trans_id)
{
	char *ret_trans_id, *tmp;
	
	ret_trans_id = tmp = vpipe("%s -b", PKG_TRANS_CMD);
	if (ret_trans_id == NULL || strlen(ret_trans_id) == 0) {
		*trans_id = '\0';
		return -1;
	}
	
	/* Nuke whitespace */
	do {
		if (*tmp == '\n' || *tmp == '\r' || *tmp == ' ')
			*tmp = '\0';
	} while (*(++tmp) != '\0');
	
	strcpy(trans_id, ret_trans_id);
	free(ret_trans_id);
	return 0;
}


/**
 * Record package addition in a transaction. Must be called after the package
 * has been added.
 * Returns 0 if successful.
 */
int
trans_add(const char *trans_id, const char *pkg_name)
{
	if (Verbose)
		printf("%s -a \"%s\" -z %s\n", PKG_TRANS_CMD, pkg_name, trans_id);
	return vsystem("%s -a \"%s\" -z %s %s", PKG_TRANS_CMD, pkg_name, trans_id,
		Verbose ? "-v" : "");
}


/**
 * Deletes a package in a transaction. Must be called before the package is
 * deleted.
 * Returns 0 if successful.
 */
int
trans_delete(const char *trans_id, const char *pkg_name)
{
	if (Verbose)
		printf("%s -d \"%s\" -z %s\n", PKG_TRANS_CMD, pkg_name, trans_id);
	return vsystem("%s -d \"%s\" -z %s %s", PKG_TRANS_CMD, pkg_name, trans_id,
		Verbose ? "-v" : "");
}


/**
 * Closes a transaction.
 */
int trans_end(const char *trans_id)
{
	if (Verbose)
		printf("%s -e -z %s\n", PKG_TRANS_CMD, trans_id);
	return vsystem("%s -e -z %s %s", PKG_TRANS_CMD, trans_id, 
		Verbose ? "-v" : "");
}
