/*-
 * Copyright (c) 2008 Ivan Voras <ivoras@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <fcntl.h>
#include <regex.h>
#include <err.h>
#include <md5.h>

#include "lib.h"


enum ACTION { ACT_ERROR, ACT_BEGIN, ACT_ADD, ACT_DELETE, ACT_END, ACT_UNDO,
	ACT_INFO, ACT_LIST };

enum TRANS_COMMAND { TC_ERROR, TC_ADD, TC_ADDBACKUP, TC_DELETE, 
	TC_BACKUPDELETE };
	
#define PACKAGE_NAME_MAX 200

struct trans_record {
	STAILQ_ENTRY(trans_record) linkage;
	enum TRANS_COMMAND	cmd;
	char		pkg_name[PACKAGE_NAME_MAX];
	char		pkg_file[PACKAGE_NAME_MAX];
	char		pkg_file_md5[MD5_DIGEST_STRING_LENGTH];
};

struct trans {
	unsigned int	id;
	time_t		time_start;
	unsigned int	n_records;
	Boolean		ended;
	STAILQ_HEAD(, trans_record) records;
};


static int argc;
static char **argv;

static void usage(void);
static void internal_trans_begin(void);
static void internal_trans_add(const char *trans_id, const char *pkg_name);
static void internal_trans_delete(const char *trans_id, const char *pkg_name);
static void internal_trans_end(const char *trans_id);
static FILE *get_trans_file(const char *trans_id, Boolean append);
static void search_rc_conf(const char *name, char **value);
static void search_rc_conf_expect(const char *name, char **value, const char *exp);
static Boolean trans_exists(const char *trans_id);
static void chdir_pkgtrans_dir(const char *trans_dir);
static void read_trans_records(FILE *f, struct trans *tr);
static void free_trans_records(struct trans *tr);
static void trans_retire(const char *trans_id);
static void trans_undo(const char *trans_id);
static int trans_undo_add(struct trans_record *t, Boolean hasbackup, const char *trans_id);
static int trans_undo_delete(struct trans_record *t, Boolean hasbackup, const char *trans_id);
static void trans_info(const char *trans_id);
static void trans_list(void);
static const char *str_trans_action(enum TRANS_COMMAND tc);


static void
search_rc_conf(const char *search_name, char **value)
{
	static regex_t re;
	static int has_re = FALSE;
	int status;
	FILE *f;
	char line[200];
	regmatch_t m[3];

	if (!has_re) {
		status = regcomp(&re, 
			"^([a-zA-Z0-9_]+)[ \t]*=\"?([a-zA-Z]+)\"?[ \t#]*",
			REG_EXTENDED);
		if (status != 0)
			errx(1, "regcomp failed: %d", status);
		has_re = TRUE;
	}

	*value = NULL;
	f = fopen("/etc/rc.conf", "r");
	while (fgets(line, 200, f) != NULL) {
		status = regexec(&re, line, 3, m, 0);
		if (status == 0) {
			char *name, *val;
			name = line + m[1].rm_so;
			val = line + m[2].rm_so;
			if (strncmp(name, search_name, m[1].rm_eo - m[1].rm_so) == 0) {
				int val_len;
				val_len = m[2].rm_eo - m[2].rm_so;
				*value = calloc(1, val_len + 1);
				strncpy(*value, val, val_len);
				break;
			}
		} 
	}
	fclose(f);
}


/**
 * Return a string description of the TRANS_COMMAND (ADD/DELETE...)
 */
static const char *
str_trans_action(enum TRANS_COMMAND tc)
{
	switch(tc) {
	case TC_ADD:
		return "ADD";
	case TC_ADDBACKUP:
		return "ADD+BACKUP";
	case TC_BACKUPDELETE:
		return "BACKUP+DELETE";
	case TC_DELETE:
		return "DELETE";
	default:
		return "Unknown";
	}
}


/**
 * Searches rc.conf for a given variable, but fails if the value for the
 * variable doesn't match the string given in 'exp' argument. The usage
 * case for this is e.g. searching if a variable is set to "YES".
 */
static void
search_rc_conf_expect(const char *name, char **value, const char *exp)
{
	search_rc_conf(name, value);
	if (*value != NULL && strcasecmp(*value, exp) != 0) {
		free(*value);
		*value = NULL;
	}
}


/**
 * Opens a transaction log file specified by the given canonical transaction ID
 * and returns its FILE* object. If 'append' argument is true, the file is
 * opened for appending only.
 */
static FILE *
get_trans_file(const char *trans_id, Boolean append)
{
	char trans_filename[PATH_MAX];
	char trans[MAX_TID_SIZE];
	int tid;
	FILE *f;

	tid = atoi(trans_id);
	make_trans_id_str(trans, tid);

	snprintf(trans_filename, PATH_MAX, "%s/%s", TRANS_DB_DIR, trans);

	if (append)
		f = fopen(trans_filename, "a");
	else
		f = fopen(trans_filename, "r+");

	if (f == NULL)
		err(1, "Cannot open transaction file '%s'", trans_filename);

	return (f);
}


/**
 * Prints usage string
 */
static void 
usage()
{
	printf("usage: %s [-b] [-e] [-a pkg] [-d pkg] [-u] [-z trans_id]\n", 
		argv[0]);
}


/**
 * Returns TRUE if the given transaction log file exists.
 */
static Boolean
trans_exists(const char *trans_id)
{
	char tpath[PATH_MAX];
	sprintf(tpath, "%s/%s", TRANS_DB_DIR, trans_id);
	if (access(tpath, R_OK) == 0)
		return TRUE;
	return FALSE;
}


/**
 * libinstall needs this function.
 */
void 
cleanup(int sig)
{
	if (sig)
		exit(1);
}


/**
 * Starts a new transaction and prints the transaction ID to stdout.
 */
static void
internal_trans_begin()
{
	DIR *d;
	struct dirent *dp;
	int this_tid;
	int n;
	char this_tids[MAX_TID_SIZE];
	char this_tid_path[PATH_MAX];
	int fd;
	FILE *f;
	time_t now;

	d = opendir(TRANS_DB_DIR);
	if (d == NULL) {
		if (mkdir(TRANS_DB_DIR, 0755) != 0)
			err(1, "Transaction database " TRANS_DB_DIR " doesn't exist and I can't create it");
		d = opendir(TRANS_DB_DIR);
		if (d == NULL)
			err(1, "Apparently it's tough to opendir(" TRANS_DB_DIR ") so I'm giving up");
	}

	this_tid = 0;
	while ((dp = readdir(d)) != NULL) {
		int tid = atoi(dp->d_name);
		if (tid > this_tid)
			this_tid = tid;
	}
	closedir(d);

	n = 0;
	while (1) {
		this_tid++;

		make_trans_id_str(this_tids, this_tid);
		snprintf(this_tid_path, PATH_MAX, "%s/%s", TRANS_DB_DIR,
			this_tids);

		fd = open(this_tid_path, O_CREAT | O_RDWR, 0644);
		if (fd >= 0)
			break;
		n++;
		if (n > 30)
			err(1, "Cannot create the transaction file '%s' "
				"(giving up after 30 attempts)",
				this_tid_path);
	}

	f = fdopen(fd, "r+");
	if (f == NULL)
		err(1, "fdopen failed");
	time(&now);

	fprintf(f, "# Transaction file created at %s", ctime(&now));
	fprintf(f, "id=%d\n", this_tid);
	fprintf(f, "time_start=%jd\n\n", (intmax_t)now);

	fclose(f);
	close(fd);
	printf("%s\n", this_tids);
}


/**
 * Record an ADD transaction.
 */
static void
internal_trans_add(const char *trans_id, const char *pkg_name)
{
	FILE *f;
	char *conf_save_pkg;
	
	if (!trans_exists(trans_id))
		errx(1, "Transaction not found: '%s'", trans_id);
	
	search_rc_conf_expect(TRANS_CONF_SAVE_ADD_PKG, &conf_save_pkg, "YES");
	
	if (!isinstalledpkg(pkg_name))
		errx(1, "'Add' transaction must be called when the "
			"package exists. (package '%s' not found)",
			pkg_name);
	
	if (conf_save_pkg != NULL) {
		int status;
		char ppath1[MAXPATHLEN];
		char pkg_md5[MD5_DIGEST_STRING_LENGTH];
		
		pkg_md5[0] = '\0';
		
		sprintf(ppath1, "%s-add", trans_id);
		chdir_pkgtrans_dir(ppath1);
		
		status = vsystem("%s -b \"%s\"", PKG_CREATE_CMD, pkg_name);
		if (status != 0)
			errx(1, "PKG_CREATE failed, code %d", status);
		
		sprintf(ppath1, "%s." PACKAGE_EXT, pkg_name);
		if (MD5File(ppath1, pkg_md5) != pkg_md5)
			err(1, "Cannot md5-hash '%s'", ppath1);
		
		f = get_trans_file(trans_id, TRUE);
		fprintf(f, "Add+Backup %s %s %s\n", pkg_name, ppath1,
			pkg_md5);
		fclose(f);
		
		free(conf_save_pkg);
	} else {
		f = get_trans_file(trans_id, TRUE);
		fprintf(f, "Add %s\n", pkg_name);
		fclose(f);
	}
}



/**
 * Record a transaction end.
 */
static void
internal_trans_end(const char *trans_id)
{
	FILE *f;

	f = get_trans_file(trans_id, TRUE);
	fprintf(f, "End.\n");
	fclose(f);
}


/**
 * Ensure that the transaction package directory exists and chdirs into it.
 */
static void
chdir_pkgtrans_dir(const char *trans_dir)
{
	DIR *d;
	
	d = opendir(TRANS_PKG_DIR);
	if (d == NULL) {
		if (mkdir(TRANS_PKG_DIR, 0755) != 0)
			err(1, "Package binary directory '%s' "
				"doesn't exist and can't be created",
				TRANS_PKG_DIR);
	} else
		closedir(d);
	
	if (chdir(TRANS_PKG_DIR) != 0)
		err(1, "Can't chdir to '%s'", TRANS_PKG_DIR);
	
	if (trans_dir != NULL) {
		d = opendir(trans_dir);
		if (d == NULL) {
			if (mkdir(trans_dir, 0755) != 0)
				err(1, "Package transaction directory '%s' "
					"doesn't exist and can't be created",
					trans_dir);
		} else
			closedir(d);
		if (chdir(trans_dir) != 0)
			err(1, "Can't chdir to '%s/%s'", TRANS_PKG_DIR,
				trans_dir);
	}
}


/**
 * Record a DELETE transaction. This function should be called BEFORE the
 * package is actually deleted.
 */
static void
internal_trans_delete(const char *trans_id, const char *pkg_name)
{
	FILE *f;
	char *conf_save_pkg;
	
	f = NULL;
	
	if (!trans_exists(trans_id))
		errx(1, "Transaction not found: '%s'", trans_id);
	if (!isinstalledpkg(pkg_name))
		errx(1, "Delete record must be created before the "
			"package is deleted (package '%s' not found)",
			pkg_name);
	
	search_rc_conf_expect(TRANS_CONF_SAVE_DEL_PKG, &conf_save_pkg, "YES");
	if (conf_save_pkg != NULL) {
		int status;
		char ppath1[MAXPATHLEN];
		char pkg_md5[MD5_DIGEST_STRING_LENGTH];
		
		pkg_md5[0] = '\0';
		sprintf(ppath1, "%s-del", trans_id);
		
		chdir_pkgtrans_dir(ppath1);
		if (Verbose)
			printf("%s -b \"%s\"", PKG_CREATE_CMD, pkg_name);
		status = vsystem("%s -b \"%s\"", PKG_CREATE_CMD, pkg_name);
		if (status != 0)
			errx(1, "PKG_CREATE failed, code %d", status);
		
		sprintf(ppath1, "%s." PACKAGE_EXT, pkg_name);
		if (MD5File(ppath1, pkg_md5) != pkg_md5)
			err(1, "Cannot md5-hash '%s'", ppath1);
		
		f = get_trans_file(trans_id, TRUE);
		fprintf(f, "Backup+Delete %s %s %s\n", pkg_name, ppath1,
			pkg_md5);
		fclose(f);
		free(conf_save_pkg);
	} else {
		f = get_trans_file(trans_id, TRUE);
		fprintf(f, "Delete %s\n", pkg_name);
		fclose(f);
	}
}


/**
 * Marks a transaction record file as "retired" by renaming it so it isn't
 * picked up any more.
 */
static void
trans_retire(const char *trans_id)
{
	char ppath1[PATH_MAX], ppath2[PATH_MAX];
	char tstr[PATH_MAX];
	time_t t;
	struct tm *lt;
	
	sprintf(ppath1, "%s/%s", TRANS_DB_DIR, trans_id);
	t = time(NULL);
	lt = localtime(&t);
	strftime(tstr, PATH_MAX, "%FT%R", lt);
	
	sprintf(ppath2, "%s/retired-%s-%s", TRANS_DB_DIR, tstr, trans_id);
	if (rename(ppath1, ppath2) != 0)
		err(1, "Cannot retire transaction '%s' to '%s'",
			trans_id, ppath2);
}


/**
 * Reads transaction file records.
 */
static void
read_trans_records(FILE *f, struct trans *tr)
{
	static Boolean has_re = FALSE;
	static regex_t re_assignment;
	static regex_t re_command;
	static regex_t re_ecommand;
	static regex_t re_statement;
	int status;
	char line[200];
	regmatch_t m[5];
	
	if (!has_re) {
		/* Simple assignment */
		if ( (status = regcomp(&re_assignment,
			"^[ \t]*([a-zA-Z0-9_]+)[ \t]*=[ \t]*([a-zA-Z0-9_+-]+)[ \t#]*",
			REG_EXTENDED)) != 0)
				errx(1, "regcomp failed: %d", status);
		/* Simple command, 1 argument */
		if ( (status = regcomp(&re_command,
			"^[ \t]*([a-zA-Z0-9+_-]+)[ \t]+([.a-zA-Z0-9_+-]+)[ \t#]*",
			REG_EXTENDED)) != 0)
				errx(1, "regcomp failed: %d", status);
		/* Extended command, 3 arguments */
		if ( (status = regcomp(&re_ecommand,
			"^[ \t]*([a-zA-Z0-9+_-]+)[ \t]+([.a-zA-Z0-9_+-]+)[ \t]+([.a-zA-Z0-9_+-]+)[ \t]+([.a-zA-Z0-9_+-]+)[ \t#]*",
			REG_EXTENDED)) != 0)
				errx(1, "regcomp failed: %d", status);
		/* Simple one-keyword statement */
		if ( (status = regcomp(&re_statement,
			"^[ \t]*([a-zA-Z0-9+_-]+)\\.[ \t#]*",
			REG_EXTENDED)) != 0)
				errx(1, "regcomp failed: %d", status);
		has_re = TRUE;
	}
	
	STAILQ_INIT(&tr->records);
	while (fgets(line, 200, f) != NULL) {
		if (line[0] == '\n' || line[0] == '#')
			continue;
		if (regexec(&re_assignment, line, 5, m, 0) == 0) {
			char *name, *val;
			
			line[m[1].rm_eo] = line[m[2].rm_eo] = '\0';
			name = line + m[1].rm_so;
			val = line + m[2].rm_so;
			if (strcasecmp(name, "id") == 0)
				tr->id = atoi(val);
			else if (strcasecmp(name, "time_start") == 0)
				tr->time_start = atol(val);
		} else if (regexec(&re_ecommand, line, 5, m, 0) == 0) {
			char *cmd, *pkg_name, *pkg_file, *pkg_file_md5;
			struct trans_record *t;
			
			line[m[1].rm_eo] = line[m[2].rm_eo] = line[m[3].rm_eo]
				= line[m[4].rm_eo]= '\0';
				
			cmd = line + m[1].rm_so;
			pkg_name = line + m[2].rm_so;
			pkg_file = line + m[3].rm_so;
			pkg_file_md5 = line + m[4].rm_so;
			if (strcasecmp(cmd, "Add+Backup") == 0) {
				t = calloc(1, sizeof *t);
				t->cmd = TC_ADDBACKUP;
				strlcpy(t->pkg_name, pkg_name, sizeof t->pkg_name);
				strlcpy(t->pkg_file, pkg_file, sizeof t->pkg_file);
				strlcpy(t->pkg_file_md5, pkg_file_md5, sizeof t->pkg_file_md5);
				STAILQ_INSERT_HEAD(&tr->records, t, linkage);
				tr->n_records++;
			} else if (strcasecmp(cmd, "Backup+Delete") == 0) {
				t = calloc(1, sizeof *t);
				t->cmd = TC_BACKUPDELETE;
				strlcpy(t->pkg_name, pkg_name, sizeof t->pkg_name);
				strlcpy(t->pkg_file, pkg_file, sizeof t->pkg_file);
				strlcpy(t->pkg_file_md5, pkg_file_md5, sizeof t->pkg_file_md5);
				STAILQ_INSERT_HEAD(&tr->records, t, linkage);
				tr->n_records++;
			} else
				errx(1, "Invalid extended transaction command: %s",
					cmd);
		} else if (regexec(&re_command, line, 5, m, 0) == 0) {
			char *cmd, *pkg_name;
			struct trans_record *t;
			
			line[m[1].rm_eo] = line[m[2].rm_eo] = '\0';
			cmd = line + m[1].rm_so;
			pkg_name = line + m[2].rm_so;
			if (strcasecmp(cmd, "Add") == 0) {
				t = calloc(1, sizeof *t);
				t->cmd = TC_ADD;
				strlcpy(t->pkg_name, pkg_name, sizeof t->pkg_name);
				STAILQ_INSERT_HEAD(&tr->records, t, linkage);
				tr->n_records++;
			} else if (strcasecmp(cmd, "Delete") == 0) {
				t = calloc(1, sizeof *t);
				t->cmd = TC_DELETE;
				strlcpy(t->pkg_name, pkg_name, sizeof t->pkg_name);
				STAILQ_INSERT_HEAD(&tr->records, t, linkage);
				tr->n_records++;
			} else
				errx(1, "Invalid simple transaction command: %s",
					cmd);
		} else if (regexec(&re_statement, line, 5, m, 0) == 0) {
			char *cmd;
			line[m[1].rm_eo] = '\0';
			cmd = line + m[1].rm_so;
			if (strcasecmp(cmd, "End") == 0)
				tr->ended = TRUE;
		} else
			errx(1, "Invalid transaction entry: '%s'", line);
	}
}


/**
 * Frees the transaction record list, but not the transaction structure itself.
 */
static void
free_trans_records(struct trans *tr)
{
	struct trans_record *t;

	while (!STAILQ_EMPTY(&tr->records)) {
		t = STAILQ_FIRST(&tr->records);
		STAILQ_REMOVE_HEAD(&tr->records, linkage);
		free(t);
	}
	bzero(tr, sizeof *tr);
}


/**
 * Undoes an ADD transaction.
 */
static int
trans_undo_add(struct trans_record *t, Boolean __unused hasbackup, const char __unused *trans_id) 
{
	if (!isinstalledpkg(t->pkg_name)) {
		if (!Force) {
			printf("ERROR: undoing requires deleting package '%s' "
				"but there is no such package installed.\n", 
				t->pkg_name);
			return (FALSE);
		}
	}
	if (vsystem("%s -z -1 %s", PKG_DELETE_CMD, t->pkg_name) != 0) {
		if (!Force) {
			printf("ERROR: pkg_delete failed.\n");
			return (FALSE);
		} else
			printf("WARNING: pkg_delete failed but forcing to go on.\n");
	}
	return (TRUE);
}


/**
 * Undoes a 'Delete' transaction.
 */
static int
trans_undo_delete(struct trans_record *t, Boolean hasbackup, const char *trans_id)
{
	int status;
	char trans_pkg_dir[PATH_MAX];
	
	if (!hasbackup) {
		char *origin;
		
		if (!Force) {
			printf("Cannot undo DELETE transaction without backup. "
				"Use -f to force install from ports.\n");
			return (FALSE);
		}
		
		/* We'll try to add the package via make install */
		
		find_origin(t->pkg_name, FALSE, &origin);
		if (origin == NULL) {
			printf("ERROR: Cannot find a package to add that "
				"matches %s\n", t->pkg_name);
			return (FALSE);
		}
		
		if (chdir(origin) != 0) {
			printf("ERROR: Cannot chdir to '%s'\n", origin);
			return (FALSE);
		}
		
		if (setenv(TRANS_ID_ENV, "-1", 1) != 0)
			errx(1, "setenv failed for '%s'", TRANS_ID_ENV);
		
		printf("*** INSTALLING PACKAGE IN '%s'\n", origin);
		
		status = vsystem("make install clean");
		if (status != 0) {
			printf("*** ERROR: make install clean failed in %s\n",
				origin);
			return (FALSE);
		}
		free(origin);
		return (TRUE);
	}
	
	/* Much simpler situation: there should be a direct backup of the
	 * package, which we'll install now. */
	
	sprintf(trans_pkg_dir, "%s/%s-del", TRANS_PKG_DIR, trans_id);
	if (chdir(trans_pkg_dir) != 0) {
		printf("ERROR: Cannot chdir to '%s'\n", trans_pkg_dir);
		return (FALSE);
	}
	
	if (isinstalledpkg(t->pkg_name)) {
		printf("WARNING: Package already exists: '%s'\n", t->pkg_name);
		return (TRUE);
	}
	
	if (!fexists(t->pkg_file)) {
		printf("ERROR: Backed-up package '%s' not found in '%s'\n",
			t->pkg_file, trans_pkg_dir);
		return (FALSE);
	}
	
	if (vsystem("%s -z -1 %s", PKG_ADD_CMD, t->pkg_file) != 0) {
		printf("ERROR: pkg_add failed\n");
		return (FALSE);
	}
	
	return (TRUE);
}


/**
 * Undoes a transaction.
 */
static void
trans_undo(const char *trans_id)
{
	FILE *f;
	struct trans tr;
	struct trans_record *t;
	
	if (!trans_exists(trans_id))
		errx(1, "Transaction not found: '%s'", trans_id);
	
	f = get_trans_file(trans_id, FALSE);
	read_trans_records(f, &tr);
	fclose(f);
	
	if (!tr.ended) {
		printf("WARNING: Transaction '%s' has not been cleanly "
			"closed.\n", trans_id);
		Verbose = Verbose ? 1 : Verbose;
	}
	
	if (Verbose) {
		int n;
		
		/* Note than undoing an 'Add' transaction means deleting
		 * packages and vice versa */
		 
		n = 0;
		printf("The following ports will be deleted:\n");
		STAILQ_FOREACH(t, &tr.records, linkage)
			if (t->cmd == TC_ADD || t->cmd == TC_ADDBACKUP)
				printf("\t%2d: %s\n", ++n, t->pkg_name);
		if (n == 0)
			printf("\tNone.\n");
		
		n = 0;
		printf("The following ports will be added:\n");
		STAILQ_FOREACH(t, &tr.records, linkage)
			if (t->cmd == TC_DELETE || t->cmd == TC_BACKUPDELETE)
				printf("\t%2d: %s\n", ++n, t->pkg_name);
		if (n == 0)
			printf("\tNone.\n");
		
		if (!y_or_n(FALSE, "Continue?")) {
			printf("Cancelling.\n");
			return;
		}
	}
	
	/* Walk the transaction records. Currently this will do the right thing
	 * as the packages are deleted and recorded in the order of
	 * dependencies. */
	STAILQ_FOREACH(t, &tr.records, linkage) {
		int ok = TRUE;
		switch(t->cmd) {
		case TC_ADD:
			ok = trans_undo_add(t, FALSE, trans_id);
			break;
		case TC_ADDBACKUP:
			ok = trans_undo_add(t, TRUE, trans_id);
			break;
		case TC_DELETE:
			ok = trans_undo_delete(t, FALSE, trans_id);
			break;
		case TC_BACKUPDELETE:
			ok = trans_undo_delete(t, TRUE, trans_id);
			break;
		default:
			errx(1, "Invalid transaction cmd %d", t->cmd);
		}
		if (!ok)
			printf("WARNING: Cannot undo %s transaction on %s\n",
				str_trans_action(t->cmd), t->pkg_name);
	}
	trans_retire(trans_id);
}


/**
 * Shows human readable information about a transaction.
 */
static void
trans_info(const char *trans_id)
{
	FILE *f;
	struct trans tr;
	struct trans_record *t;
	char trans_id_str[MAX_TID_SIZE];
	
	if (!trans_exists(trans_id))
		errx(1, "Transaction not found: '%s'", trans_id);
	
	f = get_trans_file(trans_id, FALSE);
	read_trans_records(f, &tr);
	fclose(f);
	
	if (!tr.ended) {
		printf("WARNING: Transaction '%s' has not been cleanly "
			"closed.\n", trans_id);
		Verbose = Verbose ? 1 : Verbose;
	}

	make_trans_id_str(trans_id_str, tr.id);
	printf("Transaction %s, started on %s\n", trans_id_str, ltime(time(NULL)));

	STAILQ_FOREACH(t, &tr.records, linkage) {
		const char *bflag;
		const char *cmd;
		
		switch(t->cmd) {
		case TC_ADD:
			cmd = "ADD";
			bflag = "";
			break;
		case TC_ADDBACKUP:
			cmd = "ADD";
			bflag = ",B";
			break;
		case TC_DELETE:
			cmd = "DEL";
			bflag = "";
			break;
		case TC_BACKUPDELETE:
			cmd = "DEL";
			bflag = ",B";
			break;
		default:
			printf("Invalid transaction record for %s", t->pkg_name);
			continue;
		}
		
		printf("%s%s %s\n", cmd, bflag, t->pkg_name);
	}
}


static void
trans_list()
{
	char trans_id_str[MAX_TID_SIZE];
	int i, max_trans_id;
	struct trans tr;
	struct trans_record *t;
	int n_trans;
	FILE *f;
	
	get_last_trans_id(trans_id_str);
	max_trans_id = atoi(trans_id_str);
	n_trans = 0;
	
	for (i = 1; i <= max_trans_id; i++) {
		int n_adds = 0, n_deletes = 0, n_backups = 0;
		
		make_trans_id_str(trans_id_str, i);
		if (!trans_exists(trans_id_str))
			continue;
		
		n_trans++;
		
		f = get_trans_file(trans_id_str, FALSE);
		read_trans_records(f, &tr);
		fclose(f);
		
		STAILQ_FOREACH(t, &tr.records, linkage) {
			switch(t->cmd) {
			case TC_ADD:
				n_adds++;
				break;
			case TC_ADDBACKUP:
				n_adds++;
				n_backups++;
				break;
			case TC_DELETE:
				n_deletes++;
				break;
			case TC_BACKUPDELETE:
				n_deletes++;
				n_backups++;
				break;
			default:
				printf("Invalid transaction record for %s", t->pkg_name);
				continue;
			}
		}
		
		printf("%s ", trans_id_str);
		if (n_adds > 0)
			printf("(%d pkgs added) ", n_adds);
		if (n_deletes > 0)
			printf("(%d pkgs removed) ", n_deletes);
		if (n_backups > 0)
			printf("(%d pkgs backed up) ", n_backups);
		printf("%s\n", ltime(tr.time_start));
		
		free_trans_records(&tr);
	}
	
	printf("%d transaction records found.\n", n_trans);
}


int 
main(int _argc, char **_argv)
{
	int ch;
	enum ACTION act = ACT_ERROR;
	char *pkg_name = NULL;
	char trans_id[MAX_TID_SIZE] = { '\0' };

	argc = _argc;
	argv = _argv;

	if (argc == 0) {
		usage();
		return (0);
	}

	while ( (ch = getopt(argc, argv, "a:bd:efhiluvz:")) != -1) {
		switch(ch) {
		case 'a':
			act = ACT_ADD;
			pkg_name = strdup(optarg);
			break;
		case 'b':
			act = ACT_BEGIN;
			break;
		case 'd':
			act = ACT_DELETE;
			pkg_name = strdup(optarg);
			break;
		case 'e':
			act = ACT_END;
			break;
		case 'f':
			Force++;
			break;
		case 'h':
			usage();
			exit(0);
		case 'i':
			act = ACT_INFO;
			break;
		case 'l':
			act = ACT_LIST;
			break;
		case 'u':
			act = ACT_UNDO;
			break;
		case 'v':
			Verbose++;
			break;
		case 'z':
			canonical_trans_id(trans_id, optarg);
			break;
		default:
			usage();
			exit(1);
		}
	}

	switch (act) {
	case ACT_BEGIN:
		internal_trans_begin();
		break;
	case ACT_END:
		if (trans_id[0] == '\0')
			errx(1, "This operation needs transaction id argument");
		internal_trans_end(trans_id);
		break;
	case ACT_ADD:
		if (trans_id[0] == '\0')
			errx(1, "This operation needs transaction id argument");
		if (pkg_name == NULL)
			errx(1, "This operation needs package name argument");
		internal_trans_add(trans_id, pkg_name);
		break;
	case ACT_DELETE:
		if (trans_id[0] == '\0')
			errx(1, "This operation needs transaction id argument");
		if (pkg_name == NULL)
			errx(1, "This operation needs package name argument");
		internal_trans_delete(trans_id, pkg_name);
		break;
	case ACT_UNDO:
		if (trans_id[0] == '\0') {
			get_last_trans_id(trans_id);
			if (trans_id[0] == '\0')
				errx(1, "No transactions to undo.");
		}
		trans_undo(trans_id);
		break;
	case ACT_INFO:
		if (trans_id[0] == '\0')
			errx(1, "This operation needs transaction id argument");
		trans_info(trans_id);
		break;
	case ACT_LIST:
		trans_list();
		break;
	default:
		usage();
		exit(0);
	}

	return (0);
}
